// IDTSurface.cpp: implementation of the IDTSurface class.

// Code by Karel Donk
// Contact me at karel@miraesoft.com for more information, or visit my homepage at
// http://www.miraesoft.com/karel/ from more codesamples.
// May be modified and used freely

#include "IDTSurface.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IDTSurface::IDTSurface()
{
	m_D3DDevice = NULL;
	m_D3DSurface = NULL;

	m_Width = 0;
	m_Height = 0;
	m_Format = D3DFMT_UNKNOWN;
}

IDTSurface::~IDTSurface()
{
	if (m_D3DSurface != NULL) 
	{
		m_D3DSurface->Release();
		m_D3DSurface = NULL;
	}

	m_Width = 0;
	m_Height = 0;
	m_Format = D3DFMT_UNKNOWN;
}

HRESULT IDTSurface::CreateSurface(UINT Width, UINT Height, D3DFORMAT PixelFormat)
{
	if (m_D3DDevice == NULL) return DT_NOT_INITIALIZED;

	if (m_D3DDevice->CreateOffscreenPlainSurface(Width, Height, PixelFormat, D3DPOOL_DEFAULT, &m_D3DSurface, NULL) != D3D_OK) return DT_ERROR;

	m_Width = Width;
	m_Height = Height;
	m_Format = PixelFormat;

	return DT_OK;
}

// Should be called to set the D3DDevice before doing anything with the surface
HRESULT IDTSurface::Initialize(LPDIRECT3DDEVICE9 pD3DDevice)
{
	if (pD3DDevice == NULL) return DT_NOT_INITIALIZED;

	m_D3DDevice = pD3DDevice;

	return DT_OK;
}

LPDIRECT3DSURFACE9 IDTSurface::GetD3DSurface()
{
	return m_D3DSurface;
}

// Crappy function which quickly loads an image into the surface, uses D3DX functions,
// and defaults to 255x255 image size.
// TODO: Write your own code to create a surface and load an image into it regardless of surface format.
// NOTE: This function currently creates the surface m_D3DSurface using the D3DPOOL_SCRATCH flag. For
// best performance, use the D3DPOOL_DEFAULT flag with a supported pixelformat for the creation of m_D3DSurface.
//
// For surfaces compatible with GDI (for use in windowed mode) use pixelformats D3DFMT_X1R5G5B5, D3DFMT_R5G6B5, 
// or D3DFMT_X8R8G8B8 (check the format of the backbuffer surface and use that for best performance).
// For surfaces compatible with the D3DPOOL_DEFAULT flag use pixelformats D3DFMT_X8R8G8B8, D3DFMT_A8R8G8B8, 
// D3DFMT_A2B10G10R10, D3DFMT_X1R5G5B5, D3DFMT_A1R5G5B5, and D3DFMT_R5G6B5
HRESULT IDTSurface::CreateSurfaceFromFile(LPCTSTR FileName, D3DFORMAT PixelFormat)
{
	if (m_D3DDevice == NULL) return DT_NOT_INITIALIZED;

	LPDIRECT3DSURFACE9 tempSurf;
	if (m_D3DDevice->CreateOffscreenPlainSurface(256, 256, PixelFormat, D3DPOOL_SCRATCH, &tempSurf, NULL) != D3D_OK) return DT_ERROR;

	D3DXIMAGE_INFO ImageInfo;
	if (D3DXLoadSurfaceFromFileA(tempSurf, NULL, NULL, FileName, NULL, D3DX_FILTER_LINEAR, D3DCOLOR_ARGB(255,0,0,0), &ImageInfo) != D3D_OK) 
	{
		tempSurf->Release();
		return DT_ERROR;
	}

	if (m_D3DDevice->CreateOffscreenPlainSurface(256, 256, PixelFormat, D3DPOOL_SCRATCH, &m_D3DSurface, NULL) != D3D_OK)
	{
		tempSurf->Release();
		return DT_ERROR;
	}

	if (D3DXLoadSurfaceFromSurface(m_D3DSurface, NULL, NULL, tempSurf, NULL, NULL, D3DX_FILTER_LINEAR, 0) != D3D_OK)
	{
		m_D3DSurface->Release();
		tempSurf->Release();
		return DT_ERROR;
	}

	tempSurf->Release();

	m_Width = 256;
	m_Height = 256;
	m_Format = PixelFormat;

	return DT_OK;
}

// And now for the beaf!
HRESULT IDTSurface::BltFast(LPDIRECT3DSURFACE9 Dest, int x, int y, RECT srcRect)
{
	if (m_D3DDevice == NULL) return DT_NOT_INITIALIZED;
	if (m_D3DSurface == NULL) return DT_NOT_INITIALIZED;
	if (Dest == NULL) return DT_ERROR;

	// Clipping
	if (srcRect.left < 0) srcRect.left = 0;
	if (srcRect.left > (int)(m_Width)) srcRect.left = (m_Width);
	if (srcRect.top < 0) srcRect.top = 0;
	if (srcRect.top > (int)(m_Height)) srcRect.top = (m_Height);

	if (srcRect.right < srcRect.left) srcRect.right = srcRect.left;
	if (srcRect.right > (int)(m_Width)) srcRect.right = (m_Width);
	if (srcRect.bottom < srcRect.top) srcRect.bottom = srcRect.top;
	if (srcRect.bottom > (int)(m_Height)) srcRect.bottom = (m_Height);

	UINT nHeight = srcRect.bottom-srcRect.top;
	UINT nWidth = srcRect.right-srcRect.left;

	D3DSURFACE_DESC sdesc;
	Dest->GetDesc(&sdesc);

	// Clipping
	if (x < 0)
	{	
		nWidth += x;
		x = 0;
	}
    if (x+nWidth > sdesc.Width)
	{
		nWidth = sdesc.Width-x;
	}
	if (y < 0)
	{
		nHeight += y;
		y = 0;
	}
	if (y+nHeight > sdesc.Height)
	{
		nHeight = sdesc.Height-y;
	}

	// Return if we don't have anything to blit
	if ((nHeight <= 0) || (nWidth <= 0)) return DT_OK;

	RECT destRect; destRect.left=x; destRect.top=y; destRect.right=x+nWidth; destRect.bottom=y+nHeight;

	// Get source and destination pixelformats
	// Optimization: Remove this and do it somewhere else where you can do it once since it makes things slower.
	DT_PIXELFMT spFmt, dpFmt;
	int ret = m_PixelFormat.GetDTPixelFormat(m_Format, &spFmt);
	if (ret != DT_OK) return ret;
	ret = m_PixelFormat.GetDTPixelFormat(sdesc.Format, &dpFmt);
	if (ret != DT_OK) return ret;

	D3DLOCKED_RECT slRect;
	D3DLOCKED_RECT dlRect;

	if (m_D3DSurface->LockRect(&slRect, &srcRect, D3DLOCK_READONLY) == D3D_OK)
	{
		if (Dest->LockRect(&dlRect, &destRect, 0) == D3D_OK)
		{
			BYTE* Source = (BYTE*)slRect.pBits;
			BYTE* Dest = (BYTE*)dlRect.pBits;

			for (UINT y1=0; y1<nHeight; y1++)
			{
				BYTE* Source2 = (BYTE*)Source;
				BYTE* Dest2 = (BYTE*)Dest;

				// If the surface formats would be the same, and we just want to do a SRCCOPY blit, we could just do the following to make it fast:
				//*(WORD*)Dest2 = *(WORD*)Source2;

				// But the following code illustrates breaking up a pixel from the source surface in ARGB components,
				// and then reassembling them on the destination surface, after color conversion if needed. With this 
				// code you can add blending effects such as translucency, multiply, overlay etc.
				// In an application with realtime animation, like a game, however you would not want to work with 
				// many different surface formats, because color conversion makes the blit slower. The best 
				// thing to do would be to use one surface format that best suites your needs.
				
				for (UINT x1=0; x1<nWidth; x1++)
				{
					if (spFmt.BitsPerPixel == 16)
					{
						WORD Sp = *(WORD*)Source2;
											
						DWORD alpha = ((Sp & spFmt.ABMask) >> spFmt.AShift);
						DWORD red = ((Sp & spFmt.RBMask) >> spFmt.RShift);
						DWORD green = ((Sp & spFmt.GBMask) >> spFmt.GShift);
						DWORD blue = ((Sp & spFmt.BBMask) >> spFmt.BShift);

						// Color conversion
						if (spFmt.Format != dpFmt.Format)
						{
							alpha = (DWORD) ((float)((float)alpha/(float)spFmt.AMaxVal) * (float)dpFmt.AMaxVal);
							red = (DWORD) ((float)((float)red/(float)spFmt.RMaxVal) * (float)dpFmt.RMaxVal);
							green = (DWORD) ((float)((float)green/(float)spFmt.GMaxVal) * (float)dpFmt.GMaxVal);
							blue = (DWORD) ((float)((float)blue/(float)spFmt.BMaxVal) * (float)dpFmt.BMaxVal);
						}

						if (dpFmt.BitsPerPixel == 16)
						{
  							WORD* Dp = (WORD*)Dest2;
							*Dp = (WORD) ((((alpha) << dpFmt.AShift) & dpFmt.ABMask) |
										(((red) << dpFmt.RShift) & dpFmt.RBMask) |
										(((green) << dpFmt.GShift) & dpFmt.GBMask) |
										(((blue) << dpFmt.BShift) & dpFmt.BBMask));
						}
						else if (dpFmt.BitsPerPixel > 16)
						{
  							DWORD* Dp = (DWORD*)Dest2;
							*Dp = (DWORD) ((((alpha) << dpFmt.AShift) & dpFmt.ABMask) |
										(((red) << dpFmt.RShift) & dpFmt.RBMask) |
										(((green) << dpFmt.GShift) & dpFmt.GBMask) |
										(((blue) << dpFmt.BShift) & dpFmt.BBMask));
						}
					}
					else if (spFmt.BitsPerPixel > 16)
					{
						DWORD Sp = *(DWORD*)Source2;
											
						DWORD alpha = ((Sp & spFmt.ABMask) >> spFmt.AShift);
						DWORD red = ((Sp & spFmt.RBMask) >> spFmt.RShift);
						DWORD green = ((Sp & spFmt.GBMask) >> spFmt.GShift);
						DWORD blue = ((Sp & spFmt.BBMask) >> spFmt.BShift);

						// Color conversion
						if (spFmt.Format != dpFmt.Format)
						{
							alpha = (DWORD) ((float)((float)alpha/(float)spFmt.AMaxVal) * (float)dpFmt.AMaxVal);
							red = (DWORD) ((float)((float)red/(float)spFmt.RMaxVal) * (float)dpFmt.RMaxVal);
							green = (DWORD) ((float)((float)green/(float)spFmt.GMaxVal) * (float)dpFmt.GMaxVal);
							blue = (DWORD) ((float)((float)blue/(float)spFmt.BMaxVal) * (float)dpFmt.BMaxVal);
						}
					
						if (dpFmt.BitsPerPixel == 16)
						{
  							WORD* Dp = (WORD*)Dest2;
							*Dp = (WORD) ((((alpha) << dpFmt.AShift) & dpFmt.ABMask) |
										(((red) << dpFmt.RShift) & dpFmt.RBMask) |
										(((green) << dpFmt.GShift) & dpFmt.GBMask) |
										(((blue) << dpFmt.BShift) & dpFmt.BBMask));
						}
						else if (dpFmt.BitsPerPixel > 16)
						{
  							DWORD* Dp = (DWORD*)Dest2;
							*Dp = (DWORD) ((((alpha) << dpFmt.AShift) & dpFmt.ABMask) |
										(((red) << dpFmt.RShift) & dpFmt.RBMask) |
										(((green) << dpFmt.GShift) & dpFmt.GBMask) |
										(((blue) << dpFmt.BShift) & dpFmt.BBMask));
						}
					}

					Dest2+=dpFmt.BitsPerPixel/8;
					Source2+=spFmt.BitsPerPixel/8;
				}

				Dest+=dlRect.Pitch;
				Source+=slRect.Pitch;
			}
		}
		else
		{
			m_D3DSurface->UnlockRect();

			return DT_ERROR;
		}

		Dest->UnlockRect();
	}
	else return DT_ERROR;

	m_D3DSurface->UnlockRect();

	return DT_OK;
}
